from concurrent.futures import Executor

from .common import _Future, copy_exception, copy_future_exception
from .wrap import CanCustomizeBind


class MapFuture(_Future):
    def __init__(self, delegate, map_fn):
        super(MapFuture, self).__init__()
        self._map_fn = map_fn
        self._set_delegate(delegate)

    def _set_delegate(self, delegate):
        with self._me_lock:
            self._delegate = delegate

        if delegate:
            self._delegate.add_done_callback(self._delegate_resolved)

    def _delegate_resolved(self, delegate):
        assert delegate is self._delegate, "BUG: called with %s, expected %s" % (
            delegate,
            self._delegate,
        )

        if delegate.cancelled():
            return

        ex = delegate.exception()
        if ex:
            copy_future_exception(delegate, self)
            return

        result = delegate.result()
        try:
            result = self._map_fn(result)
        except Exception:
            copy_exception(self)
            return

        try:
            self._on_mapped(result)
        except Exception:
            copy_exception(self)

    def _on_mapped(self, result):
        self.set_result(result)

    def set_result(self, result):
        with self._me_lock:
            super(MapFuture, self).set_result(result)
        self._me_invoke_callbacks()

    def set_exception(self, exception):
        with self._me_lock:
            super(MapFuture, self).set_exception(exception)
        self._me_invoke_callbacks()

    def set_exception_info(self, exception, traceback):
        # For python2 compat.
        # pylint: disable=no-member
        with self._me_lock:
            super(MapFuture, self).set_exception_info(exception, traceback)
        self._me_invoke_callbacks()

    def running(self):
        with self._me_lock:
            if self.done():
                return False
            return self._delegate and (
                self._delegate.running() or self._delegate.done()
            )

    def _me_cancel(self):
        return self._delegate.cancel()


class MapExecutor(CanCustomizeBind, Executor):
    """An executor which delegates to another executor while mapping
    output values through a given function.
    """

    _FUTURE_CLASS = MapFuture

    def __init__(self, delegate, fn, logger=None):
        """
        Arguments:
            delegate (~concurrent.futures.Executor):
                an executor to which callables will be submitted
            fn (callable):
                a callable applied to transform returned values
            logger (~logging.Logger):
                a logger used for messages from this executor
        """
        self._delegate = delegate
        self._fn = fn

    def shutdown(self, wait=True):
        self._delegate.shutdown(wait)

    def submit(self, fn, *args, **kwargs):
        """Submit a callable.

        The returned future will have its output value transformed by the
        map function passed to this executor.  If that map function raises
        an exception, the future will fail with that exception."""
        inner_f = self._delegate.submit(fn, *args, **kwargs)
        return self._FUTURE_CLASS(inner_f, self._fn)



# -*- coding: utf-8 -*-

import weakref
from threading import Lock
import logging
from concurrent.futures import Future

from .base import chain_cancel
from ..common import copy_future_exception
from .check import ensure_futures


LOG = logging.getLogger("more_executors.futures")


class BoolCallback(object):
    AWAITING_FIRST = 1
    AWAITING_SECOND = 2
    STOP = 3

    def __init__(self, output, f1, f2):
        self.state = self.AWAITING_FIRST
        self.output = output
        self.lock = Lock()
        self.f1 = weakref.ref(f1)
        self.f2 = weakref.ref(f2)

    def get_cancel_future(self, completed_future):
        f1 = self.f1()
        f2 = self.f2()

        if f1 and completed_future is f1 and f2:
            return f2
        if f2 and completed_future is f2 and f1:
            return f1

    def update_state(self, future, have_result):
        raise NotImplementedError()  # pragma: no cover

    def __call__(self, future):
        if self.output.cancelled():
            return

        exception = future.exception() if not future.cancelled() else None
        result = None
        set_result = None
        have_result = False
        set_exception = None
        cancel_future = None
        if not exception and not future.cancelled():
            result = future.result()
            have_result = True

        with self.lock:
            (set_result, set_exception, cancel_future) = self.update_state(
                future, have_result
            )

        if set_result:
            self.output.set_result(result)
        if set_exception:
            copy_future_exception(future, self.output)
        if cancel_future:
            cancelled = cancel_future.cancel()
            LOG.debug(
                "cancel %s in callback for %s: %s", cancel_future, future, cancelled
            )


class OrCallback(BoolCallback):
    def update_state(self, future, have_result):
        set_result = None
        set_exception = None
        cancel_future = None

        if self.state == self.AWAITING_FIRST:
            if have_result and future.result():
                self.state = self.STOP
                set_result = True
                cancel_future = self.get_cancel_future(future)
                LOG.debug(
                    "f_or: set result on %s to %s from left=%s",
                    self.output,
                    future.result(),
                    future,
                )
            else:
                self.state = self.AWAITING_SECOND
        elif self.state == self.AWAITING_SECOND:
            self.state = self.STOP
            if future.cancelled():
                cancel_future = self.output
            elif have_result:
                set_result = True
                LOG.debug(
                    "f_or: set result on %s to %s from right=%s",
                    self.output,
                    future.result(),
                    future,
                )
            else:
                set_exception = True

        return (set_result, set_exception, cancel_future)


class AndCallback(BoolCallback):
    def update_state(self, future, have_result):
        set_result = None
        set_exception = None
        cancel_future = None

        if self.state == self.AWAITING_SECOND:
            self.state = self.STOP
            if future.cancelled():
                cancel_future = self.output
            elif future.exception():
                set_exception = True
            else:
                set_result = True
                LOG.info(
                    "f_and: set %s result %s due completion of %s",
                    self.output,
                    future.result(),
                    future,
                )
        elif self.state == self.AWAITING_FIRST:
            if future.cancelled():
                cancel_future = self.output
                self.state = self.STOP
            elif future.exception():
                set_exception = True
                cancel_future = self.get_cancel_future(future)
                self.state = self.STOP
            elif not future.result():
                set_result = True
                cancel_future = self.get_cancel_future(future)
                self.state = self.STOP
            else:
                # This side gave a truthy result so await the other
                self.state = self.AWAITING_SECOND

        return (set_result, set_exception, cancel_future)


@ensure_futures
def f_or(f, *fs):
    """Boolean ``OR`` over a number of futures.

    Signature: :code:`Future<A>[, Future<B>[, ...]] ⟶ Future<A|B|...>`

    Arguments:
        f (~concurrent.futures.Future)
            Any future
        fs (~concurrent.futures.Future)
            Any futures

    Returns:
        :class:`~concurrent.futures.Future`
            A future resolved from the inputs using ``OR`` semantics:

            - Resolved with the earliest true value returned by an input
              future, if any.
            - Otherwise, resolved with the latest false value or exception
              returned by the input futures.

    .. versionadded:: 1.19.0
    """
    if not fs:
        return f

    f1 = f
    f2 = f_or(fs[0], *fs[1:])

    out = Future()
    chain_cancel(out, f1)
    chain_cancel(out, f2)
    callback = OrCallback(out, f1, f2)
    f1.add_done_callback(callback)
    f2.add_done_callback(callback)
    return out


@ensure_futures
def f_and(f, *fs):
    """Boolean ``AND`` over a number of futures.

    Signature: :code:`Future<A>[, Future<B>[, ...]] ⟶ Future<A|B|...>`

    Arguments:
        f (~concurrent.futures.Future)
            Any future
        fs (~concurrent.futures.Future)
            Any futures

    Returns:
        :class:`~concurrent.futures.Future`
            A future resolved from the inputs using ``AND`` semantics:

            - Resolved with the latest value returned by an input
              future, if all futures are resolved with true values.
            - Otherwise, resolved with the earliest false value or exception
              returned by the input futures.

    .. versionadded:: 1.19.0
    """
    if not fs:
        return f

    f1 = f
    f2 = f_and(fs[0], *fs[1:])

    out = Future()
    chain_cancel(out, f1)
    chain_cancel(out, f2)
    callback = AndCallback(out, f1, f2)
    f1.add_done_callback(callback)
    f2.add_done_callback(callback)
    return out



from ._impl.futures import (
    f_apply,
    f_or,
    f_and,
    f_return,
    f_return_error,
    f_return_cancelled,
    f_zip,
    f_map,
    f_flat_map,
    f_traverse,
    f_sequence,
    f_nocancel,
    f_timeout,
)


__all__ = [
    "f_apply",
    "f_or",
    "f_and",
    "f_return",
    "f_return_error",
    "f_return_cancelled",
    "f_zip",
    "f_map",
    "f_flat_map",
    "f_traverse",
    "f_sequence",
    "f_nocancel",
    "f_timeout",
]



from .map import MapFuture, MapExecutor


class FlatMapFuture(MapFuture):
    def __init__(self, *args, **kwargs):
        self.__flattened = False
        super(FlatMapFuture, self).__init__(*args, **kwargs)

    def _on_mapped(self, result):
        if self.__flattened:
            return super(FlatMapFuture, self)._on_mapped(result)

        # Result *must* be a future.
        # We'd crash in set_delegate if not.
        # Let's raise this more helpful exception prior to that.
        if not callable(getattr(result, "add_done_callback", None)):
            raise TypeError(
                "FlatMapExecutor's function did not return a Future!\n"
                "  Function: %s\n"
                "  Returned: %s" % (repr(self._map_fn), repr(result))
            )

        self.__flattened = True
        self._map_fn = lambda x: x
        self._set_delegate(result)


class FlatMapExecutor(MapExecutor):
    """An executor which delegates to another executor while mapping
    output values through a given future-producing function.

    This executor behaves like :class:`~more_executors.map.MapExecutor`,
    except that the given mapping function must return instances of
    :class:`~concurrent.futures.Future`, and the mapped future is
    flattened into the future returned from this executor.
    This allows chaining multiple future-producing functions into a single
    future.

    - If the map function returns a :class:`~concurrent.futures.Future`, the
      result/exception of that future will be propagated to the future returned
      by this executor.
    - If the map function returns any other type, the returned future will fail
      with a :class:`TypeError`.
    - If the map function raises an exception, the returned future will fail
      with that exception.

    .. versionadded: 1.12.0
    """

    _FUTURE_CLASS = FlatMapFuture



# -*- coding: utf-8 -*-

from .base import wrap, f_return
from .apply import f_apply
from .check import ensure_future


@ensure_future
def f_map(future, fn):
    """Map the output value of a future using the given function.

    Signature: :code:`Future<A>, fn<A⟶B> ⟶ Future<B>`

    Arguments:
        future (~concurrent.futures.Future)
            Any future.
        fn (callable)
            Any callable.

    Returns:
        :class:`~concurrent.futures.Future`
            A future resolved with the returned value of :code:`fn(future.result())`,
            or with the exception raised by :obj:`future` or :obj:`fn`.

    .. versionadded:: 1.19.0
    """
    future_fn = f_return(fn)
    return f_apply(future_fn, future)


@ensure_future
def f_flat_map(future, fn):
    """Map the output value of a future using the given future-returning function.

    Like :meth:`f_map`, except that the mapping function must return a future,
    and that future will be flattened into the output value (avoiding a nested future).

    Signature: :code:`Future<A>, fn<A⟶Future<B>> ⟶ Future<B>`

    Arguments:
        future (~concurrent.futures.Future)
            Any future.
        fn (callable)
            Any callable which returns a future.

    Returns:
        :class:`~concurrent.futures.Future`
            A future equivalent to that returned by :obj:`fn`, or resolved with an exception
            if :obj:`future` or :obj:`fn` failed.

    .. versionadded:: 1.19.0
    """
    return wrap(future).with_flat_map(fn)()



"""A library of composable Python executors.

- [Documentation](https://rohanpm.github.io/more-executors/)
- [Source](https://github.com/rohanpm/more-executors)
- [PyPI](https://pypi.python.org/pypi/more-executors)

This library is intended for use with the
[`concurrent.futures`](https://docs.python.org/3/library/concurrent.futures.html)
module.  It includes a collection of `Executor` implementations in order to
extend the behavior of `Future` objects.

Compatible with Python 2.6, 2.7 and 3.x.

This documentation was built from an unknown revision.
"""
from ._impl.executors import Executors

__all__ = ["Executors"]



from concurrent.futures import Executor, Future

from .common import copy_exception
from .wrap import CanCustomizeBind


class SyncExecutor(CanCustomizeBind, Executor):
    """An executor which immediately invokes all submitted callables."""

    def __init__(self, logger=None):
        """
        Parameters:
            logger (~logging.Logger):
                a logger used for messages from this executor
        """
        super(SyncExecutor, self).__init__()

    def submit(self, fn, *args, **kwargs):
        """Immediately invokes `fn(*args, **kwargs)` and returns a future
        with the result (or exception)."""
        future = Future()
        try:
            result = fn(*args, **kwargs)
            future.set_result(result)
        except Exception:
            copy_exception(future)
        return future



from concurrent.futures import Executor
from threading import RLock
import logging

from .wrap import CanCustomizeBind


class CancelOnShutdownExecutor(CanCustomizeBind, Executor):
    """An executor which delegates to another executor and cancels all
    futures when the executor is shut down.

    This class is useful in conjunction with executors having custom cancel
    behavior, such as :class:`~more_executors.poll.PollExecutor`.
    """

    def __init__(self, delegate, logger=None):
        """
        Parameters:

            delegate (~concurrent.futures.Executor):
                executor to which callables will be submitted

            logger (~logging.Logger):
                a logger used for messages from this executor
        """
        self._log = logger if logger else logging.getLogger("CancelOnShutdownExecutor")
        self._delegate = delegate
        self._futures = set()
        self._lock = RLock()
        self._shutdown = False

    def shutdown(self, wait=True):
        """Shut down the executor.

        All futures created by this executor which have not yet been completed
        will have :meth:`~concurrent.futures.Future.cancel` invoked.

        Note that there is no guarantee that the cancel will succeed, and only a single
        attempt is made to cancel any future.
        """
        with self._lock:
            if self._shutdown:
                return
            self._shutdown = True
            futures = self._futures.copy()

        for f in futures:
            cancel = f.cancel()
            self._log.debug("Cancel %s: %s", f, cancel)

        self._delegate.shutdown(wait)

    def submit(self, fn, *args, **kwargs):
        with self._lock:
            if self._shutdown:
                raise RuntimeError("Cannot submit after shutdown")
            future = self._delegate.submit(fn, *args, **kwargs)
            self._futures.add(future)
            future.add_done_callback(self._futures.discard)
        return future



from concurrent.futures import Executor
from threading import RLock, Thread, Event
import weakref
import logging

from .common import _Future, MAX_TIMEOUT, copy_exception, copy_future_exception
from .wrap import CanCustomizeBind
from .helpers import executor_loop


class PollFuture(_Future):
    def __init__(self, delegate, executor):
        super(PollFuture, self).__init__()
        self._delegate = delegate
        self._executor = executor
        self._delegate.add_done_callback(self._delegate_resolved)
        self.add_done_callback(self._clear_executor)

    def _delegate_resolved(self, delegate):
        assert delegate is self._delegate, "BUG: called with %s, expected %s" % (
            delegate,
            self._delegate,
        )

        if delegate.cancelled():
            return
        elif delegate.exception():
            copy_future_exception(delegate, self)
        else:
            self._executor._register_poll(self, self._delegate)

    def _clear_delegate(self):
        with self._me_lock:
            self._delegate = None

    @classmethod
    def _clear_executor(cls, future):
        future._executor._deregister_poll(future)
        future._executor = None

    def set_result(self, result):
        with self._me_lock:
            if self.done():
                return
            super(PollFuture, self).set_result(result)
        self._me_invoke_callbacks()

    def set_exception(self, exception):
        with self._me_lock:
            # We can't get here because we always try set_exception_info first,
            # which succeeds if done() is true, in which case we won't fall
            # back to set_exception.
            # if self.done():
            #    return
            super(PollFuture, self).set_exception(exception)
        self._me_invoke_callbacks()

    def set_exception_info(self, exception, traceback):
        # For python2 compat.
        # pylint: disable=no-member
        with self._me_lock:
            if self.done():
                return
            super(PollFuture, self).set_exception_info(exception, traceback)
        self._me_invoke_callbacks()

    def running(self):
        with self._me_lock:
            if self._delegate:
                return self._delegate.running()
        # If delegate is removed, we're now polling or done.
        return False

    def _me_cancel(self):
        if self._delegate and not self._delegate.cancel():
            return False
        executor = self._executor
        return executor and executor._run_cancel_fn(self)


class PollDescriptor(object):
    """Represents an unresolved :class:`~concurrent.futures.Future`.

    The poll function used by :class:`PollExecutor` will be
    invoked with a list of PollDescriptor objects.
    """

    def __init__(self, future, result):
        self.__future = future
        self.__result = result

    @property
    def result(self):
        """The result from the delegate executor's future, which should be used to
        drive the poll."""
        return self.__result

    def yield_result(self, result):
        """The poll function can call this function to make the future yield the given result.

        Arguments:
            result (object):
                a result to be returned by the future associated with this descriptor
        """

        self.__future.set_result(result)

    def yield_exception(self, exception, traceback=None):
        """The poll function can call this function to make the future raise the given exception.

        Arguments:
           exception (Exception):
               An exception to be returned or raised by the future associated with this
               descriptor
           traceback (traceback):
               An optional associated traceback. This argument only has an effect on python 2.x,
               where exception objects do not include a traceback.
        """

        copy_exception(self.__future, exception, traceback)


class PollExecutor(CanCustomizeBind, Executor):
    """Instances of `PollExecutor` submit callables to a delegate executor
    and resolve the returned futures via a provided poll function.

    A cancel function may also be provided to perform additional processing
    when a returned future is cancelled.
    """

    def __init__(
        self, delegate, poll_fn, cancel_fn=None, default_interval=5.0, logger=None
    ):
        """
        Parameters:

            delegate (~concurrent.futures.Executor):
                executor to which callables will be submitted

            poll_fn (callable):
                a `poll function`_ used to decide when futures should be resolved

            cancel_fn (callable):
                a `cancel function`_ invoked when future cancel is required

            default_interval (float):
                default interval between polls (in seconds)

            logger (~logging.Logger):
                a logger used for messages from this executor
        """
        self._log = logger if logger else logging.getLogger("PollExecutor")
        self._delegate = delegate
        self._default_interval = default_interval
        self._poll_fn = poll_fn
        self._cancel_fn = cancel_fn
        self._poll_descriptors = []
        self._poll_event = Event()

        poll_event = self._poll_event
        self_ref = weakref.ref(self, lambda _: poll_event.set())

        self._poll_thread = Thread(
            name="PollExecutor", target=_poll_loop, args=(self_ref,)
        )
        self._poll_thread.daemon = True
        self._shutdown = False
        self._lock = RLock()

        self._poll_thread.start()

    def submit(self, fn, *args, **kwargs):
        delegate_future = self._delegate.submit(fn, *args, **kwargs)
        return PollFuture(delegate_future, self)

    def _register_poll(self, future, delegate_future):
        descriptor = PollDescriptor(future, delegate_future.result())
        with self._lock:
            self._poll_descriptors.append((future, descriptor))
            future._clear_delegate()
            self._poll_event.set()

    def _deregister_poll(self, future):
        with self._lock:
            self._poll_descriptors = [
                (f, d) for (f, d) in self._poll_descriptors if f is not future
            ]

    def _run_cancel_fn(self, future):
        if not self._cancel_fn:
            # no cancel function => no veto of cancel
            return True

        descriptor = [d for (f, d) in self._poll_descriptors if f is future]
        if not descriptor:
            # no record of this future => no veto of cancel.
            # we can get here if the future is already done
            # or if polling hasn't started yet
            return True

        assert len(descriptor) == 1, "Too many poll descriptors for %s" % future

        descriptor = descriptor[0]

        try:
            return self._cancel_fn(descriptor.result)
        except Exception:
            self._log.exception(
                "Exception during cancel on %s/%s", future, descriptor.result
            )
            return False

    def _run_poll_fn(self):
        with self._lock:
            descriptors = [d for (_, d) in self._poll_descriptors]

        try:
            return self._poll_fn(descriptors)
        except Exception as e:
            self._log.debug("Poll function failed", exc_info=True)
            # If poll function fails, then every future
            # depending on the poll also immediately fails.
            [d.yield_exception(e) for d in descriptors]

    def shutdown(self, wait=True):
        self._shutdown = True
        self._poll_event.set()
        self._delegate.shutdown(wait)
        if wait:
            self._log.debug("Join poll thread...")
            self._poll_thread.join(MAX_TIMEOUT)
            self._log.debug("Joined poll thread.")


@executor_loop
def _poll_loop(executor_ref):
    while True:
        executor = executor_ref()
        if not executor:
            break

        if executor._shutdown:
            break

        executor._log.debug("Polling...")

        next_sleep = executor._run_poll_fn()
        if not (isinstance(next_sleep, int) or isinstance(next_sleep, float)):
            next_sleep = executor._default_interval

        executor._log.debug("Sleeping...")

        poll_event = executor._poll_event
        del executor

        poll_event.wait(next_sleep)
        poll_event.clear()



from .wrapped import CustomizableThreadPoolExecutor, CustomizableProcessPoolExecutor
from .bind import BoundCallable
from .map import MapExecutor
from .flat_map import FlatMapExecutor
from .retry import RetryExecutor
from .poll import PollExecutor
from .timeout import TimeoutExecutor
from .throttle import ThrottleExecutor
from .cancel_on_shutdown import CancelOnShutdownExecutor
from .sync import SyncExecutor
from .asyncio import AsyncioExecutor


class Executors(object):
    """Convenience methods for creating executors.

    This class produces wrapped executors which may be customized
    by use of the `with_*` methods, as in the following example:

    >>> Executors.thread_pool(max_workers=4).with_retry().with_map(lambda x: x*2)

    Produces a thread pool executor which will retry on failure
    and multiply all output values by 2."""

    @classmethod
    def bind(cls, executor, fn):
        """Bind a synchronous callable to an executor.

        If the callable returns a future, consider using :meth:`flat_bind` instead.

        Arguments:
            executor (~concurrent.futures.Executor): an executor
            fn (callable): any function or callable

        Returns:
            callable:
                A new callable which, when invoked, will submit `fn` to `executor` and return
                the resulting future.

                This returned callable provides the `Executors.with_*` methods, which may be
                chained to further customize the behavior of the callable.

        .. versionadded:: 1.13.0
        """
        return BoundCallable(executor, fn)

    @classmethod
    def flat_bind(cls, executor, fn):
        """Bind an asynchronous callable to an executor.

        This convenience method should be used in preference to :meth:`bind`
        when the bound callable returns a future, in order to avoid a nested
        future in the returned value. It is equivalent to:

        >>> bind(fn).with_flat_map(lambda future: future)

        Arguments:
            executor (~concurrent.futures.Executor): an executor
            fn (callable): any function or callable which returns a future

        Returns:
            callable:
                A new callable which, when invoked, will submit `fn` to `executor` and return
                the resulting (flattened) future.

                This returned callable provides the `Executors.with_*` methods, which may be
                chained to further customize the behavior of the callable.

        .. versionadded:: 1.16.0
        """
        return cls.bind(executor, fn).with_flat_map(lambda f: f)

    @classmethod
    def thread_pool(cls, *args, **kwargs):
        """Create a thread pool executor.

        Returns:
            ~concurrent.futures.ThreadPoolExecutor:
                a new executor, initialized with the given arguments
        """
        return CustomizableThreadPoolExecutor(*args, **kwargs)

    @classmethod
    def process_pool(cls, *args, **kwargs):
        """Create a process pool executor.

        Returns:
            ~concurrent.futures.ProcessPoolExecutor:
                a new executor, initialized with the given arguments
        """
        return CustomizableProcessPoolExecutor(*args, **kwargs)

    @classmethod
    def sync(cls, *args, **kwargs):
        """Creates a new synchronous executor.

        Returns:
            ~more_executors.sync.SyncExecutor:
                a new synchronous executor

        Submitted functions will be immediately invoked on the calling thread."""
        return SyncExecutor(*args, **kwargs)

    @classmethod
    def _customize(cls, delegate, executor_class, *args, **kwargs):
        if isinstance(delegate, BoundCallable):
            executor = delegate._BoundCallable__executor
            bound_fn = delegate._BoundCallable__fn
            new_executor = executor_class(executor, *args, **kwargs)
            return cls.bind(new_executor, bound_fn)

        return executor_class(delegate, *args, **kwargs)

    @classmethod
    def with_retry(cls, executor, *args, **kwargs):
        """
        Returns:
            ~more_executors.retry.RetryExecutor:
                a new executor which will retry callables on failure
        """
        return cls._customize(executor, RetryExecutor, *args, **kwargs)

    @classmethod
    def with_map(cls, executor, *args, **kwargs):
        """
        Returns:
            ~more_executors.map.MapExecutor:
                a new executor which will transform results through the given function
        """
        return cls._customize(executor, MapExecutor, *args, **kwargs)

    @classmethod
    def with_flat_map(cls, executor, *args, **kwargs):
        """
        Returns:
            ~more_executors.flat_map.FlatMapExecutor:
                a new executor which will transform results through the given
                :class:`~concurrent.futures.Future`-providing function

        .. versionadded:: 1.12.0
        """
        return cls._customize(executor, FlatMapExecutor, *args, **kwargs)

    @classmethod
    def with_poll(cls, executor, *args, **kwargs):
        """
        Returns:
            ~more_executors.poll.PollExecutor:
                a new executor which produces polled futures.

                Submitted callables will have their output passed into
                a poll function.
        """
        return cls._customize(executor, PollExecutor, *args, **kwargs)

    @classmethod
    def with_timeout(cls, executor, *args, **kwargs):
        """
        Returns:
            ~more_executors.timeout.TimeoutExecutor:
                a new executor which will attempt to cancel any futures if they've
                not completed within the given timeout.

        .. versionadded:: 1.7.0
        """
        return cls._customize(executor, TimeoutExecutor, *args, **kwargs)

    @classmethod
    def with_throttle(cls, executor, *args, **kwargs):
        """
        Returns:
            ~more_executors.throttle.ThrottleExecutor:
                a new executor which enforces a limit on the number of concurrently
                pending futures.

        .. versionadded:: 1.9.0
        """
        return cls._customize(executor, ThrottleExecutor, *args, **kwargs)

    @classmethod
    def with_cancel_on_shutdown(cls, executor, *args, **kwargs):
        """
        Returns:
            ~more_executors.cancel_on_shutdown.CancelOnShutdownExecutor:
                a new executor which attempts to cancel any pending futures when
                the executor is shut down.
        """
        return cls._customize(executor, CancelOnShutdownExecutor, *args, **kwargs)

    @classmethod
    def with_asyncio(cls, executor, *args, **kwargs):
        """
        Returns:
            ~more_executors.asyncio.AsyncioExecutor:
                a new executor which returns :class:`asyncio.Future` instances
                rather than :class:`concurrent.futures.Future` instances, i.e. may be used
                with the `await` keyword and coroutines.

        .. note::
            Since the other executors from :class:`Executors` class are designed
            for use with :class:`concurrent.futures.Future`, if an executor is being
            configured using chained `with_*` methods, this must be the last method called.

        .. note::
            Only usable for Python >= 3.5.

        .. versionadded:: 1.7.0
        """
        return cls._customize(executor, AsyncioExecutor, *args, **kwargs)



from concurrent.futures import Executor
from threading import Event, Thread, Lock
from collections import namedtuple
import weakref
import logging

from monotonic import monotonic

from .map import MapFuture
from .common import MAX_TIMEOUT
from .wrap import CanCustomizeBind
from .helpers import executor_loop

LOG = logging.getLogger("TimeoutExecutor")

Job = namedtuple("Job", ["future", "delegate_future", "deadline"])


class TimeoutFuture(MapFuture):
    def __init__(self, delegate):
        super(TimeoutFuture, self).__init__(delegate, lambda x: x)


class TimeoutExecutor(CanCustomizeBind, Executor):
    """An executor which delegates to another executor while applying
    a timeout to each returned future.

    For any futures returned by this executor, if the future hasn't
    completed approximately within `timeout` seconds of its creation,
    an attempt will be made to cancel the future.

    Note that only a single attempt is made to cancel any future, and there
    is no guarantee that this will succeed.

    .. versionadded:: 1.7.0
    """

    def __init__(self, delegate, timeout, logger=None):
        """
        Parameters:

            delegate (~concurrent.futures.Executor):
                an executor to which callables are submitted

            timeout (float):
                timeout (in seconds) after which :meth:`concurrent.futures.Future.cancel()`
                will be invoked on any generated future which has not completed

            logger (~logging.Logger):
                a logger used for messages from this executor
        """
        self._log = logger if logger else LOG
        self._delegate = delegate
        self._timeout = timeout
        self._shutdown = False
        self._jobs = []
        self._jobs_lock = Lock()
        self._jobs_write = Event()

        event = self._jobs_write
        self_ref = weakref.ref(self, lambda _: event.set())

        self._job_thread = Thread(
            name="TimeoutExecutor", target=self._job_loop, args=(self_ref,)
        )
        self._job_thread.daemon = True
        self._job_thread.start()

    def submit(self, fn, *args, **kwargs):
        return self.submit_timeout(self._timeout, fn, *args, **kwargs)

    def submit_timeout(self, timeout, fn, *args, **kwargs):
        """Like :code:`submit(fn, *args, **kwargs)`, but uses the specified
        timeout rather than this executor's default.

        .. versionadded:: 1.19.0
        """
        delegate_future = self._delegate.submit(fn, *args, **kwargs)
        future = TimeoutFuture(delegate_future)
        future.add_done_callback(self._on_future_done)
        job = Job(future, delegate_future, monotonic() + timeout)
        with self._jobs_lock:
            self._jobs.append(job)
        self._jobs_write.set()
        return future

    def shutdown(self, wait=True):
        self._log.debug("shutdown")
        self._shutdown = True
        self._jobs_write.set()
        self._delegate.shutdown(wait)
        if wait:
            self._job_thread.join(MAX_TIMEOUT)

    def _partition_jobs(self):
        pending = []
        overdue = []
        now = monotonic()
        for job in self._jobs:
            if job.future.done():
                self._log.debug("Discarding job for completed future: %s", job)
            elif job.deadline < now:
                overdue.append(job)
            else:
                pending.append(job)
        return (pending, overdue)

    def _on_future_done(self, future):
        self._log.debug("Waking thread for %s", future)
        self._jobs_write.set()

    def _do_cancel(self, job):
        self._log.debug("Attempting cancel: %s", job)
        cancel_result = job.future.cancel()
        self._log.debug("Cancel of %s resulted in %s", job, cancel_result)

    @classmethod
    @executor_loop
    def _job_loop(cls, executor_ref):
        while True:
            (event, wait_time) = cls._job_loop_iter(executor_ref())
            if not event:
                break
            event.wait(wait_time)
            event.clear()

    @classmethod
    def _job_loop_iter(cls, executor):
        if not executor:
            LOG.debug("Executor was collected")
            return (None, None)

        if executor._shutdown:
            executor._log.debug("Executor was shut down")
            return (None, None)

        executor._log.debug("job loop")

        with executor._jobs_lock:
            (pending, overdue) = executor._partition_jobs()
            executor._jobs = pending

        executor._log.debug("jobs: %s overdue, %s pending", len(overdue), len(pending))

        for job in overdue:
            executor._do_cancel(job)

        wait_time = None
        if pending:
            earliest = min([job.deadline for job in pending])
            wait_time = max(earliest - monotonic(), 0)

        executor._log.debug("Wait until %s", wait_time)
        return (executor._jobs_write, wait_time)



# -*- coding: utf-8 -*-

from ..map import MapFuture
from .check import ensure_future


class NoCancelFuture(MapFuture):
    def cancel(self):
        return False


@ensure_future
def f_nocancel(future):
    """Wrap a future to block cancellation.

    Signature: :code:`Future<X> ⟶ Future<X>`

    Arguments:
        future (~concurrent.futures.Future)
            Any future.

    Returns:
        :class:`~concurrent.futures.Future`
            A wrapped version of :obj:`future` which cannot be cancelled.

    .. versionadded:: 1.19.0
    """
    return NoCancelFuture(future, lambda x: x)



# -*- coding: utf-8 -*-

from .base import f_return, chain_cancel
from .map import f_map, f_flat_map
from .check import ensure_futures


@ensure_futures
def f_zip(*fs):
    """Create a new future holding the return values of any number of input futures.

    Signature: :code:`Future<A>[, Future<B>[, ...]] ⟶ Future<A[, B[, ...]]>`

    Arguments:
        fs (~concurrent.futures.Future)
            Any number of futures.

    Returns:
        :class:`~concurrent.futures.Future` of :class:`tuple`
            A future holding the returned values of all input futures as a tuple.
            The returned tuple has the same length and order as the input futures.

            Alternatively, a future raising an exception, if any input futures raised
            an exception.

    .. versionadded:: 1.19.0
    """
    if not fs:
        return f_return(())

    f = fs[0]
    rest = f_zip(*fs[1:])

    def prepender(f_result):
        return f_map(rest, lambda result: tuple([f_result] + list(result)))

    out = f_flat_map(f, prepender)
    chain_cancel(out, f)
    chain_cancel(out, rest)
    return out



import logging
from functools import wraps

LOG = logging.getLogger(__name__)


def executor_loop(fn):
    @wraps(fn)
    def out(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except RuntimeError as error:
            if "cannot schedule new futures after" in str(error):
                LOG.debug("Ignoring error due to interpreter shutdown", exc_info=1)
                return
            raise

    return out



from concurrent.futures import Executor
from threading import RLock, Thread, Event
import logging
import weakref

from monotonic import monotonic

from .common import _Future, MAX_TIMEOUT, copy_future_exception
from .wrap import CanCustomizeBind
from .helpers import executor_loop


class RetryPolicy(object):
    """Instances of this class may be supplied to :class:`RetryExecutor`
    to customize the retry behavior.

    This base class will never retry.  See :class:`ExceptionRetryPolicy` for
    a general-purpose implementation."""

    def should_retry(self, attempt, future):
        """
        Parameters:
            attempt (int): number of times the future has been attempted; starts counting at 1
            future (~concurrent.futures.Future): a completed future

        Returns:
            bool:
                True if and only if a future should be retried.
        """
        return False

    def sleep_time(self, attempt, future):
        """
        Parameters:
            attempt (int): number of times the future has been attempted; starts counting at 1
            future (~concurrent.futures.Future): a completed future

        Returns:
            float:
                The amount of time (in seconds) to delay before the next
                attempt at running a future.
        """
        return 0


class ExceptionRetryPolicy(RetryPolicy):
    """Retries on any exceptions under the given base class(es),
    up to a fixed number of attempts, with an exponential
    backoff between each attempt."""

    def __init__(self, **kwargs):
        """
        Parameters:
            max_attempts (int): maximum number of times a callable should be attempted
            exponent (float): exponent used for backoff, e.g. 2.0 will result in a delay
                              which doubles between each attempt
            sleep (float): base value for delay between attempts in seconds, e.g. 1.0
                           will delay by one second between first two attempts
            max_sleep (float): maximum delay between attempts, in seconds
            exception_base (type, list(type)): future will be retried if (and only if)
                            it raises an exception inheriting from one of these classes

        All parameters are optional; reasonable defaults apply when omitted.
        """
        self._max_attempts = kwargs.get("max_attempts", 3)
        self._exponent = kwargs.get("exponent", 2.0)
        self._sleep = kwargs.get("sleep", 1.0)
        self._max_sleep = kwargs.get("max_sleep", 120)
        self._exception_base = kwargs.get("exception_base", Exception)
        if isinstance(self._exception_base, type):
            self._exception_base = [self._exception_base]

    def should_retry(self, attempt, future):
        exception = future.exception()
        if not exception:
            return False
        if attempt >= self._max_attempts:
            return False
        for klass in self._exception_base:
            if isinstance(exception, klass):
                return True
        return False

    def sleep_time(self, attempt, future):
        return min(self._sleep * (self._exponent ** attempt), self._max_sleep)


class RetryJob(object):
    def __init__(
        self,
        policy,
        delegate_future,
        future,
        attempt,
        when,
        fn,
        args,
        kwargs,
        old_delegate=None,
    ):
        self.policy = policy
        self.delegate_future = delegate_future
        self.future = future
        self.attempt = attempt
        self.when = when
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.stop_retry = False
        self.old_delegate = old_delegate


class RetryFuture(_Future):
    def __init__(self, executor):
        super(RetryFuture, self).__init__()
        self.delegate_future = None
        self._executor = executor
        self.add_done_callback(self._clear_executor)

    def running(self):
        with self._me_lock:
            if self.done():
                return False
            # Note that if we are not "done", but the delegate future is "done",
            # we consider ourselves as running - this should mean the delegate
            # future completed but callbacks haven't finished yet (and in fact,
            # we might decide to retry)
            if self.delegate_future:
                return self.delegate_future.running() or self.delegate_future.done()
        return False

    def _clear_delegate(self):
        with self._me_lock:
            self.delegate_future = None

    @classmethod
    def _clear_executor(cls, future):
        with future._me_lock:
            future._executor = None

    def __terminate_via(self, method, *args, **kwargs):
        with self._me_lock:
            self._clear_delegate()
            method(*args, **kwargs)
        self._me_invoke_callbacks()

    def set_result(self, result):
        self.__terminate_via(super(RetryFuture, self).set_result, result)

    def set_exception(self, exception):
        self.__terminate_via(super(RetryFuture, self).set_exception, exception)

    def set_exception_info(self, exception, traceback):
        # For python2 compat.
        # pylint: disable=no-member
        self.__terminate_via(
            super(RetryFuture, self).set_exception_info, exception, traceback
        )

    def _me_cancel(self):
        executor = self._executor
        return executor and executor._cancel(self)


class RetryExecutor(CanCustomizeBind, Executor):
    """An executor which delegates to another executor while adding
    implicit retry behavior.

    - Callables are submitted to the delegate executor, and may be
      submitted more than once if retries are required.

    - The callables may be submitted to that executor from a different
      thread than the calling thread.

    - Cancelling is supported if the delegate executor allows it.

    - Cancelling between retries is always supported.

    - Attempting to cancel a future prevents any more retries, regardless
      of whether the cancel succeeds.

    - The returned futures from this executor are only resolved or
      failed once the callable either succeeded, or all retries
      were exhausted.  This includes activation of the done callback.

    - This executor is thread-safe.
    """

    def __init__(self, delegate, retry_policy=None, logger=None, **kwargs):
        """
        Parameters:

            delegate (~concurrent.futures.Executor):
                executor to which callables will be submitted

            retry_policy (RetryPolicy):
                policy used to determine when futures shall be retried; if omitted,
                an :class:`ExceptionRetryPolicy` is used, supplied with keyword
                arguments to this constructor

            logger (~logging.Logger):
                a logger used for messages from this executor
        """
        self._log = logger if logger else logging.getLogger("RetryExecutor")
        self._delegate = delegate
        self._default_retry_policy = retry_policy or ExceptionRetryPolicy(**kwargs)
        self._jobs = []
        self._submit_event = Event()

        event = self._submit_event
        self_ref = weakref.ref(self, lambda _: event.set())
        self._submit_thread = Thread(
            name="RetryExecutor", target=_submit_loop, args=(self_ref,)
        )
        self._submit_thread.daemon = True

        self._shutdown = False
        self._lock = RLock()

        self._submit_thread.start()

    def shutdown(self, wait=True):
        self._log.debug("Shutting down.")
        self._shutdown = True
        self._wake_thread()
        self._delegate.shutdown(wait)
        if wait:
            self._log.debug("Waiting for thread")
            self._submit_thread.join(MAX_TIMEOUT)
        self._log.debug("Shutdown complete")

    def submit(self, fn, *args, **kwargs):
        return self.submit_retry(self._default_retry_policy, fn, *args, **kwargs)

    def submit_retry(self, retry_policy, fn, *args, **kwargs):
        """Submit a callable with a specific retry policy.

        Parameters:
            retry_policy (RetryPolicy): a policy which is used for this call only
        """
        future = RetryFuture(self)

        job = RetryJob(retry_policy, None, future, 0, monotonic(), fn, args, kwargs)
        self._append_job(job)

        # Let the submit thread know it should wake up to check for new jobs
        self._wake_thread()

        self._log.debug("Returning future %s", future)
        return future

    def _wake_thread(self):
        self._submit_event.set()

    def _get_next_job(self):
        # Find and return the next job to be handled, if any.
        # This means a job with when < now, or with stop_retry == True,
        # or if there's none, then the job with the minimum value of when.
        min_job = None
        now = monotonic()
        for job in self._jobs:
            if job.delegate_future:
                # It's already running, skip
                continue
            if job.stop_retry:
                # We've been requested to stop retrying this job,
                # and we can handle that immediately
                return job
            if job.when <= now:
                # job is overdue, just do it
                return job
            elif not min_job:
                min_job = job
            elif job.when < min_job.when:
                min_job = job
        return min_job

    def _submit_now(self, job):
        # Pop job since we'll replace it.
        # We need to hold the lock for the entire duration so that other
        # threads won't see _jobs between our removal and re-add of the job
        with job.future._me_lock:
            with self._lock:
                self._pop_job(job)

                # We need the future's lock now too, because someone could
                # call cancel after this check and before we submit.
                if job.future.done():
                    self._log.debug(
                        "future done %s - not submitting to delegate", job.future
                    )
                    return

                delegate_future = self._delegate.submit(job.fn, *job.args, **job.kwargs)
                job.future.delegate_future = delegate_future

                new_job = RetryJob(
                    job.policy,
                    delegate_future,
                    job.future,
                    job.attempt + 1,
                    None,
                    job.fn,
                    job.args,
                    job.kwargs,
                )
                self._append_job(new_job)
                self._log.debug("Submitted: %s", new_job)

        delegate_future.add_done_callback(self._delegate_callback)
        self._wake_thread()

    def _pop_job(self, job):
        with self._lock:
            for idx, pending in enumerate(self._jobs):
                if pending is job:
                    return self._jobs.pop(idx)

    def _append_job(self, job):
        with self._lock:
            self._jobs.append(job)

    def _retry(self, job, sleep_time):
        self._log.debug("Will retry: %s", job)

        with self._lock:
            self._pop_job(job)
            new_job = RetryJob(
                job.policy,
                None,
                job.future,
                job.attempt,
                monotonic() + sleep_time,
                job.fn,
                job.args,
                job.kwargs,
                # The old delegate future is retained in case we need
                # to propagate its exception later
                old_delegate=job.delegate_future,
            )
            new_job.stop_retry = job.stop_retry
            self._append_job(new_job)

        self._wake_thread()

    def _cancel(self, future):
        found_job = None

        with self._lock:
            for idx, job in enumerate(self._jobs):
                if job.future is future:
                    self._log.debug("Try cancel: %s", job)

                    if not job.delegate_future:
                        self._log.debug("Successful cancel - no delegate: %s", job)
                        self._jobs.pop(idx)
                        return True

                    found_job = job

                    # Whether or not we can successfully cancel, the request to cancel
                    # means that we don't want to retry any more.
                    found_job.stop_retry = True

                    break

        # This shouldn't be possible.
        # - Future holds a lock on itself, and has checked that it's not already done
        # - The only other path for removing a job is in delegate_callback, but the
        #   job is only removed *after* set_result/set_exception which would wait
        #   for the future's lock.
        assert found_job, "Cancel called on orphan %s" % future

        self._log.debug("Try cancel delegate: %s", found_job)

        if found_job.delegate_future.cancel():
            self._log.debug("Successful cancel: %s", found_job)
            future._clear_delegate()
            # Don't remove from _jobs here,
            # the callback attached to delegate_future was expected
            # to take care of that
            return True

        self._log.debug("Could not cancel: %s", found_job)

        # Let the submit thread wake up and find that we've set stop_retry
        self._wake_thread()

        return False

    def _delegate_callback(self, delegate_future):
        assert delegate_future.done(), "BUG: callback invoked while future not done!"

        self._log.debug("Callback activated for %s", delegate_future)

        found_job = None
        for job in self._jobs[:]:
            if job.delegate_future == delegate_future:
                found_job = job
                break

        # Callbacks are only installed after a job is added, and this is
        # the only place a job with a delegate associated will be removed,
        # thus it should not be possible for a job to be missing.
        assert found_job, "BUG: no job associated with delegate %s" % delegate_future

        if delegate_future.cancelled():
            # nothing to do, retrying on cancel is not allowed
            self._log.debug("Delegate was cancelled: %s", delegate_future)
            return

        if found_job.stop_retry:
            should_retry = False
        else:
            should_retry = found_job.policy.should_retry(
                found_job.attempt, delegate_future
            )

        if should_retry:
            sleep_time = found_job.policy.sleep_time(found_job.attempt, delegate_future)
            self._retry(found_job, sleep_time)
            return

        self._log.debug("Finalizing %s", found_job)

        # OK, it won't be retried.  Resolve the future.
        copy_future(delegate_future, found_job.future)

        self._pop_job(found_job)

        self._log.debug("Finalized %s", found_job)


def copy_future(f1, f2):
    exception = f1.exception()
    if exception:
        result = None
    else:
        result = f1.result()

    # OK, it won't be retried.  Resolve the future.
    if exception:
        copy_future_exception(f1, f2)
    else:
        f2.set_result(result)


@executor_loop
def _submit_loop(executor_ref):
    # Runs in a separate thread continuously submitting to the delegate
    # executor until no jobs are ready, or waiting until next job is ready
    while True:
        executor = executor_ref()
        if not executor:
            break

        if executor._shutdown:
            break

        executor._log.debug("_submit_loop iter")

        with executor._lock:
            job = executor._get_next_job()

        if not job:
            executor._log.debug("No jobs at all. Waiting...")
            event = executor._submit_event
            del executor
            _submit_wait(event)
            continue

        if job.stop_retry:
            executor._log.debug("Discarding job due to cancel: %s", job)
            executor._pop_job(job)
            copy_future(job.old_delegate, job.future)
            continue

        now = monotonic()
        if job.when <= now:
            # Can submit immediately and check for next job
            executor._submit_now(job)
            continue

        # There is nothing to submit immediately.
        # Sleep until either:
        # - reaching the time of the nearest job, or...
        # - woken up by condvar
        delta = job.when - now
        executor._log.debug("No ready job.  Waiting: %s", delta)
        event = executor._submit_event
        del executor
        del job
        _submit_wait(event, delta)


def _submit_wait(event, timeout=None):
    event.wait(timeout)
    event.clear()



from functools import update_wrapper

from .wrap import CanCustomize


class BoundCallable(CanCustomize, object):
    def __init__(self, executor, fn):
        self.__executor = executor
        self.__fn = fn

        try:
            update_wrapper(self, fn)
        except AttributeError:
            # Update wrapper if we can, but not fatal if we can't
            pass

    def __call__(self, *args, **kwargs):
        return self.__executor.submit(self.__fn, *args, **kwargs)



from concurrent.futures import Executor
from threading import Event, Thread, Lock, Semaphore
from collections import namedtuple, deque
from functools import partial
import logging
import weakref

from .common import MAX_TIMEOUT
from .wrap import CanCustomizeBind
from .map import MapFuture
from .helpers import executor_loop


class ThrottleFuture(MapFuture):
    def __init__(self, executor):
        self._executor = executor
        super(ThrottleFuture, self).__init__(delegate=None, map_fn=lambda x: x)
        self.add_done_callback(self._clear_executor)

    def _me_cancel(self):
        if self._delegate:
            return self._delegate.cancel()
        executor = self._executor
        return executor and executor._do_cancel(self)

    @classmethod
    def _clear_executor(cls, future):
        future._executor = None


ThrottleJob = namedtuple("ThrottleJob", ["future", "fn", "args", "kwargs"])


class ThrottleExecutor(CanCustomizeBind, Executor):
    """An executor which delegates to another executor while enforcing
    a limit on the number of futures running concurrently.

    - Callables are submitted to the delegate executor, from a different
      thread than the calling thread.

    - Where `count` is used to initialize this executor, if there
      are already `count` futures submitted to the delegate executor and not
      yet :meth:`~concurrent.futures.Future.done`, additional callables will
      be queued and only submitted to the delegate executor once there are
      less than `count` futures in progress.

    .. versionadded:: 1.9.0
    """

    def __init__(self, delegate, count, logger=None):
        """
        Parameters:
            delegate (~concurrent.futures.Executor):
                an executor to which callables will be submitted

            count (int):
                maximum number of concurrently running futures

            logger (~logging.Logger):
                a logger used for messages from this executor
        """
        self._log = logger if logger else logging.getLogger("ThrottleExecutor")
        self._delegate = delegate
        self._to_submit = deque()
        self._lock = Lock()
        self._event = Event()
        self._sem = Semaphore(count)
        self._shutdown = False

        event = self._event
        self_ref = weakref.ref(self, lambda _: event.set())

        self._thread = Thread(
            name="ThrottleExecutor", target=_submit_loop, args=(self_ref,)
        )
        self._thread.daemon = True
        self._thread.start()

    def submit(self, fn, *args, **kwargs):
        out = ThrottleFuture(self)
        job = ThrottleJob(out, fn, args, kwargs)
        with self._lock:
            self._to_submit.append(job)
            self._log.debug("Enqueued: %s", job)
        self._event.set()
        return out

    def shutdown(self, wait=True):
        self._log.debug("Shutting down")
        self._shutdown = True
        self._delegate.shutdown(wait)
        self._event.set()
        if wait:
            self._thread.join(MAX_TIMEOUT)

    def _do_submit(self, job):
        delegate_future = self._delegate.submit(job.fn, *job.args, **job.kwargs)
        self._log.debug("Submitted %s yielding %s", job, delegate_future)

        delegate_future.add_done_callback(
            partial(self._delegate_future_done, self._log, self._sem, self._event)
        )
        job.future._set_delegate(delegate_future)

    def _do_cancel(self, future):
        with self._lock:
            for job in self._to_submit:
                if job.future is future:
                    self._to_submit.remove(job)
                    self._log.debug("Cancelled %s", job)
                    return True
        self._log.debug("Could not find for cancel: %s", future)
        return False

    @classmethod
    def _delegate_future_done(cls, log, sem, event, future):
        # Whenever an inner future completes, one more execution slot becomes
        # available, and the thread should wake up in case there's something to
        # be submitted
        log.debug("Delegate future done: %s", future)
        sem.release()
        event.set()


def _submit_loop_iter(executor):
    if not executor:
        return

    if executor._shutdown:
        return

    to_submit = []
    with executor._lock:
        while executor._to_submit:
            if not executor._sem.acquire(False):
                executor._log.debug("Throttled")
                break
            job = executor._to_submit.popleft()
            executor._log.debug("Will submit: %s", job)
            to_submit.append(job)

        executor._log.debug(
            "Submitting %s, throttling %s", len(to_submit), len(executor._to_submit)
        )

    for job in to_submit:
        executor._do_submit(job)

    return executor._event


@executor_loop
def _submit_loop(executor_ref):
    while True:
        event = _submit_loop_iter(executor_ref())
        if not event:
            break

        event.wait()
        event.clear()



from functools import wraps


def ensure_futures(f):
    @wraps(f)
    def new_fn(*args, **kwargs):
        to_check = list(args)
        to_check.extend(kwargs.values())

        for arg in to_check:
            if not is_future(arg):
                raise TypeError(
                    "%s() called with non-future value: %s" % (f.__name__, repr(arg))
                )

        return f(*args, **kwargs)

    return new_fn


def ensure_future(f):
    @wraps(f)
    def new_fn(*args, **kwargs):
        arg = None
        if args:
            arg = args[0]
        else:
            arg = kwargs.get("future")

        if not is_future(arg):
            raise TypeError(
                "%s() called with non-future value: %s" % (f.__name__, repr(arg))
            )

        return f(*args, **kwargs)

    return new_fn


def is_future(f):
    return "add_done_callback" in dir(f)



from .apply import f_apply
from .bool import f_or, f_and
from .base import f_return, f_return_error, f_return_cancelled
from .zip import f_zip
from .map import f_map, f_flat_map
from .sequence import f_traverse, f_sequence
from .nocancel import f_nocancel
from .timeout import f_timeout

__all__ = [
    "f_apply",
    "f_or",
    "f_and",
    "f_return",
    "f_return_error",
    "f_return_cancelled",
    "f_zip",
    "f_map",
    "f_flat_map",
    "f_traverse",
    "f_sequence",
    "f_nocancel",
    "f_timeout",
]



# -*- coding: utf-8 -*-

from .base import wrap
from .check import ensure_futures


# for wrapping arguments.
# This value means an argument came from *args rather than **kwargs
ARGS = object()


@ensure_futures
def f_apply(future_fn, *future_args, **future_kwargs):
    """Call a function, where the function, its arguments and return value
    are all provided by futures.

    Signature: :code:`Future<fn<A[,B[,...]]⟶R>>, Future<A>[, Future<B>[, ...]] ⟶ Future<R>`

    Arguments:
        future_fn (:class:`~concurrent.futures.Future` of :class:`callable`)
            A future returning a function to be applied.
        future_args (~concurrent.futures.Future)
            Futures holding positional arguments for the function.
        future_kwargs (~concurrent.futures.Future)
            Futures holding keyword arguments for the function.

    Returns:
        ~concurrent.futures.Future:
            A future holding the returned value of the applied function.

    .. versionadded:: 1.19.0
    """
    wrapped_args = _wrap_args(*future_args, **future_kwargs)
    return _wrapped_f_apply(future_fn, wrapped_args)


def _wrap_args(*future_args, **future_kwargs):
    out = list()
    for arg in future_args:
        out.append((ARGS, arg))
    for key, value in future_kwargs.items():
        out.append((key, value))
    return out


def _wrapped_f_apply(future_fn, future_args):
    if not future_args:
        return wrap(future_fn).with_map(lambda fn: fn())()

    future_key_and_x = future_args[0]
    (key, future_x) = future_key_and_x
    future_args = future_args[1:]

    # future_fn takes multiple arguments.
    # Create a new equivalent future function which takes one less arg
    def fn_runner(fn, x):
        def out(*args, **kwargs):
            args = list(args)
            kwargs = kwargs.copy()
            if key is ARGS:
                args.insert(0, x)
            else:
                kwargs[key] = x
            return fn(*args, **kwargs)

        return out

    next_future_fn = wrap(future_x).with_flat_map(
        lambda x: wrap(future_fn).with_map(lambda fn: fn_runner(fn, x))()
    )()

    return _wrapped_f_apply(next_future_fn, future_args)



from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

from .wrap import CanCustomizeBind


class CustomizableThreadPoolExecutor(CanCustomizeBind, ThreadPoolExecutor):
    pass


class CustomizableProcessPoolExecutor(CanCustomizeBind, ProcessPoolExecutor):
    pass



from setuptools import setup, find_packages


def get_description():
    return "A library of composable Python executors and futures"


def get_long_description():
    text = open("README.md").read()

    # The README starts with the same text as "description",
    # which makes sense, but on PyPI causes same text to be
    # displayed twice.  So let's strip that.
    return text.replace(get_description() + ".\n\n", "", 1)


def get_install_requires():
    return open("requirements.txt").readlines()


setup(
    name="more-executors",
    version="2.1.2",
    author="Rohan McGovern",
    author_email="rohan@mcgovern.id.au",
    packages=find_packages(exclude=["tests", "tests.*"]),
    url="https://github.com/rohanpm/more-executors",
    license="GNU General Public License",
    description=get_description(),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=get_install_requires(),
    project_urls={
        "Changelog": "https://github.com/rohanpm/more-executors/blob/master/CHANGELOG.md",
        "Documentation": "https://rohanpm.github.io/more-executors/",
    },
)



# -*- coding: utf-8 -*-

from concurrent.futures import Future

from ..executors import Executors
from ..common import copy_exception


def f_return(x=None):
    """Return a future which provides the value `x`.

    Signature: :code:`A ⟶ Future<A>`

    Arguments:
        x
            A value to be returned

    Returns:
        :class:`~concurrent.futures.Future` of :obj:`x`
            A future immediately resolved with the value :obj:`x`.

    .. versionadded:: 1.19.0
    .. versionchanged:: 2.1.0
        value now defaults to :code:`None`
    """
    return Executors.sync().submit(lambda: x)


def f_return_error(x, traceback=None):
    """Return a future which returns/raises the exception `x`.

    Arguments:

        x (Exception)
            An exception to be returned/raised.

        traceback (traceback)
            An optional traceback associated with the exception.

            This argument should only be provided on Python 2.x.
            It will be used in the value returned from
            :meth:`concurrent.futures.Future.exception_info`,
            which only exists on Python 2.

    Returns:
        :class:`~concurrent.futures.Future` of :obj:`x`
            A future immediately resolved with the exception :obj:`x`.

    .. versionadded:: 1.19.0
    """

    f = Future()
    copy_exception(f, x, traceback)
    return f


def f_return_cancelled():
    """
    Returns:
        :class:`~concurrent.futures.Future`
            A future which is cancelled.

    .. versionadded:: 1.19.0
    """
    f = Future()
    f.cancel()
    f.set_running_or_notify_cancel()
    return f


def chain_cancel(f_outer, f_inner):
    # attempt to cancel f_inner when f_outer is cancelled
    f_outer.add_done_callback(lambda f: f_inner.cancel() if f.cancelled() else False)


def wrap(f):
    return Executors.sync().flat_bind(lambda: f)



import sys
from concurrent.futures import Future
from threading import RLock
import logging


LOG = logging.getLogger("more_executors._Future")

# This value should be used for any blocking waits likely to be invoked
# from the main thread, where blocking forever is technically appropriate.
#
# The reason for this is that, in Python 2.x, a blocking wait with no
# timeout (such as a thread join) is entirely uninterruptible, always
# retrying on EINTR, which can easily lead to a process not responding
# to anything but SIGKILL.
#
# Providing a timeout value - no matter what it is - causes the wait to
# become interruptible, which is desirable.
#
# This value is an arbitrary choice.  100 years ought to be enough for anyone :)
MAX_TIMEOUT = 60 * 60 * 24 * 365 * 100


class _Future(Future):
    # Need to reimplement some of the Future class.
    # We must hold callbacks ourselves and not let the parent class handle them.
    # This is necessary for locking: we need to handle some locking ourselves,
    # but we must NOT have the callbacks invoked while our lock is held.
    # (Note: an alternative would be to access concurrent.futures private variables)
    def __init__(self):
        super(_Future, self).__init__()
        self._me_done_callbacks = []
        self._me_lock = RLock()

    def _me_invoke_callbacks(self):
        for callback in self._me_done_callbacks:
            try:
                callback(self)
            except Exception:
                LOG.exception("exception calling callback for %r", self)

        # Drop references to the callbacks once no longer required,
        # so that futures don't keep other objects alive longer than needed
        self._me_done_callbacks = []

    def add_done_callback(self, fn):
        # Overrides function from parent; intentionally does not call super,
        # so we own the callbacks ourself
        with self._me_lock:
            if not self.done():
                self._me_done_callbacks.append(fn)
                return
        # Already done -> call it directly
        fn(self)

    def cancel(self):
        with self._me_lock:
            if self.cancelled():
                return True
            if self.done():
                return False
            if not self._me_cancel():
                return False
            out = super(_Future, self).cancel()
            if out:
                self.set_running_or_notify_cancel()
        if out:
            self._me_invoke_callbacks()
        return out

    def _me_cancel(self):
        raise NotImplementedError(
            "BUG: override this method in subclasses!"
        )  # pragma: no cover


def copy_future_exception(f1, f2):
    if "exception_info" in dir(f1):
        (exception, traceback) = f1.exception_info()
    else:
        (exception, traceback) = (f1.exception(), None)

    copy_exception(f2, exception, traceback)


def copy_exception(future, exception=None, traceback=None):
    exc_info = sys.exc_info()
    if exception is None:
        exception = exc_info[1]
    if traceback is None:
        traceback = exc_info[2]

    try:
        future.set_exception_info(exception, traceback)
    except AttributeError:
        future.set_exception(exception)



# -*- coding: utf-8 -*-

from concurrent.futures import Future

from .zip import f_zip
from .map import f_map

from ..common import copy_exception


def f_sequence(futures):
    """Transform a list of futures into a future containing a list.

    Signature: :code:`list<Future<X>> ⟶ Future<list<X>>`

    Arguments:
        futures (iterable of :class:`~concurrent.futures.Future`)
            A list or other iterable of futures.

    Returns:
        :class:`~concurrent.futures.Future` of :class:`list`
            A future resolved with either:

            - a list holding the output value of each input future
            - or an exception, if any input future raised an exception

    .. versionadded:: 1.19.0
    """
    return f_traverse(lambda x: x, futures)


def f_traverse(fn, xs):
    """Traverse over an iterable calling a future-returning function,
    and return a future holding the returned values as a list.

    Signature: :code:`fn<A⟶Future<B>>, iterable<A> ⟶ Future<list<B>>`

    Arguments:
        fn (callable)
            A unary function returning a future.
        xs (iterable)
            An iterable to be traversed.

    Returns:
        :class:`~concurrent.futures.Future` of :class:`list`
            A future resolved with either:

            - a list holding the resolved output values of :obj:`fn`
            - or an exception, if :obj:`fn` or any future produced by :obj:`fn` failed

    .. versionadded:: 1.19.0
    """
    try:
        futures = [fn(x) for x in xs]
    except Exception:
        future = Future()
        copy_exception(future)
        return future

    zipped = f_zip(*futures)
    return f_map(zipped, list)



# -*- coding: utf-8 -*-

import weakref
from threading import Lock

from more_executors import Executors

from .check import ensure_future

LOCK = Lock()
EXECUTOR_REF = None


@ensure_future
def f_timeout(future, timeout):
    """Wrap a future to cancel it after a timeout is reached.

    Signature: :code:`Future<X>, float ⟶ Future<X>`

    Arguments:
        future (~concurrent.futures.Future)
            Any future.
        timeout (float)
            A timeout to apply to the future, in seconds.

    Returns:
        :class:`~concurrent.futures.Future`
            A wrapped version of :obj:`future` which may be cancelled if the
            future has not completed within :obj:`timeout` seconds.

            Note: only a single attempt is made to cancel the future, and there
            is no guarantee that the cancel will succeed.

    .. versionadded:: 1.19.0
    """
    return timeout_executor().submit_timeout(timeout, lambda: future)


def timeout_executor():
    global EXECUTOR_REF  # pylint: disable=global-statement
    with LOCK:
        executor = EXECUTOR_REF and EXECUTOR_REF()
        if not executor:
            executor = Executors.sync().with_flat_map(lambda x: x).with_timeout(None)
            EXECUTOR_REF = weakref.ref(executor)
        return executor



class CanBind(object):
    def bind(self, *args, **kwargs):
        from .executors import Executors

        return Executors.bind(self, *args, **kwargs)

    def flat_bind(self, *args, **kwargs):
        from .executors import Executors

        return Executors.flat_bind(self, *args, **kwargs)


class CanCustomize(object):
    def with_retry(self, *args, **kwargs):
        from .executors import Executors

        return Executors.with_retry(self, *args, **kwargs)

    def with_map(self, *args, **kwargs):
        from .executors import Executors

        return Executors.with_map(self, *args, **kwargs)

    def with_flat_map(self, *args, **kwargs):
        from .executors import Executors

        return Executors.with_flat_map(self, *args, **kwargs)

    def with_poll(self, *args, **kwargs):
        from .executors import Executors

        return Executors.with_poll(self, *args, **kwargs)

    def with_timeout(self, *args, **kwargs):
        from .executors import Executors

        return Executors.with_timeout(self, *args, **kwargs)

    def with_throttle(self, *args, **kwargs):
        from .executors import Executors

        return Executors.with_throttle(self, *args, **kwargs)

    def with_cancel_on_shutdown(self, *args, **kwargs):
        from .executors import Executors

        return Executors.with_cancel_on_shutdown(self, *args, **kwargs)

    def with_asyncio(self, *args, **kwargs):
        from .executors import Executors

        return Executors.with_asyncio(self, *args, **kwargs)


class CanCustomizeBind(CanBind, CanCustomize):
    pass



from concurrent.futures import Executor
import asyncio


class AsyncioExecutor(Executor):
    """An executor which delegates to another executor while converting
    returned futures into instances of :class:`asyncio.Future`.

    Note that since this class produces :mod:`asyncio` rather than :mod:`concurrent.futures`
    future objects, AsyncioExecutor instances themselves cannot be used
    as a delegate executor of other executor instances within this library.

    .. versionadded:: 1.7.0
    """

    def __init__(self, delegate, loop=None, logger=None):
        """
        Parameters:

            delegate (~concurrent.futures.Executor):
                executor to which callables will be submitted

            loop (~asyncio.AbstractEventLoop):
                asyncio event loop used to wrap futures; if omitted, the default
                event loop is used.

            logger (~logging.Logger):
                a logger used for messages from this executor
        """
        self._delegate = delegate
        self._loop = loop

    def submit(self, fn, *args, **kwargs):
        return self.submit_with_loop(self._loop, fn, *args, **kwargs)

    def submit_with_loop(self, loop, fn, *args, **kwargs):
        """Submit a callable with the specified event loop.

        Parameters:
            loop (~asyncio.AbstractEventLoop):
                asyncio event loop used to wrap futures

            fn (callable):
                callable to be submitted

        Returns:
            asyncio.Future:
                a future for the given callable
        """
        if not loop:
            loop = asyncio.get_event_loop()
        future = self._delegate.submit(fn, *args, **kwargs)
        return asyncio.wrap_future(future, loop=loop)

    def shutdown(self, wait=True):
        self._delegate.shutdown(wait)
